import json

class NDayInterval:
    """
    表示返回N天前的日期，可用于SQL中与 ">=" 运算符配合，筛选最近N天的数据。
    例如: NDayInterval(7) => DATE_SUB(NOW(), INTERVAL 7 DAY)
    用法示例：WHERE date_column >= DATE_SUB(NOW(), INTERVAL 7 DAY)
    """
    def __init__(self, days):
        self.days = days
    def __str__(self):
        return f"DATE_SUB(NOW(), INTERVAL {self.days} DAY)"


def _validate_param_value(param_value, field_name):
    """
    校验参数值是否为numpy类型，如果是则抛出异常；
    如果是Dict类型，自动执行json.dumps转换为字符串
    
    :param param_value: 需要校验的参数值
    :param field_name: 参数字段名，用于错误信息
    :return: 处理后的参数值
    :raises: TypeError - 当参数值为numpy类型时，或Dict类型json.dumps失败时
    """
    # 检查是否为numpy类型
    param_type = type(param_value).__name__
    if param_type.startswith('numpy'):
        raise TypeError(f"字段 '{field_name}' 的参数值类型为 {param_type}，numpy类型数据无法直接写入数据库，请先转换")
    
    # 如果是Dict类型，自动执行json.dumps
    if isinstance(param_value, dict):
        try:
            return json.dumps(param_value, ensure_ascii=False)
        except (TypeError, ValueError) as e:
            raise TypeError(f"字段 '{field_name}' 的Dict类型参数值：{param_value} 无法转换为JSON字符串: {str(e)}")
    
    return param_value


def build_where_clause( conditions ) :
    """
    构造WHERE子句和对应的参数列表
    
    该方法用于根据提供的条件字典构建SQL的WHERE子句，支持两种条件格式：
    1. 简单值：自动生成等值比较条件
    2. 元组格式：自定义比较运算符和值
    
        :param conditions: WHERE条件，格式为字典，每个键为字段名，每个值可以是:
        - 简单值: 默认使用 = 比较，如 {'name': '张三', 'age': 25}
        - 元组: (比较运算符, 值) 如 {'age': ('>', 18)}, {'name': ('LIKE', '%张%')}
        - 支持的比较运算符包括: =, !=, <>, >, >=, <, <=, LIKE, NOT LIKE, IN, NOT IN
    
    :return: tuple - (where_clause, params)
        - where_clause: str - 构建好的WHERE子句（不包含WHERE关键字），条件之间用AND连接
        - params: list - 对应的参数值列表，用于防止SQL注入
    
    :raises: ValueError - 当元组格式长度不为2时
    :raises: TypeError - 当参数值为numpy类型时
    
    :example:
        >>> conditions = {'name': '张三', 'age': ('>', 18)}
        >>> clause, params = build_where_clause(conditions)
        >>> print(clause)  # 输出: name = %s AND age > %s
        >>> print(params)  # 输出: ['张三', 18]

        >>> from lazy_mysql.tools.where_clause import NDayInterval
        >>> conditions = {'order_dateTime': ('>=', NDayInterval(7))}
        >>> clause, params = build_where_clause(conditions)
        >>> print(clause)  # 输出: order_dateTime >= DATE_SUB(NOW(), INTERVAL 7 DAY)
        >>> print(params)  # 输出: []

        >>> conditions = {'status': ('IN', [1, 2, 3]), 'create_time': ('>=', '2023-01-01')}
        >>> clause, params = build_where_clause(conditions)
        >>> print(clause)  # 输出: status IN (%s, %s, %s) AND create_time >= %s
        >>> print(params)  # 输出: [1, 2, 3, '2023-01-01']
    """
    if not conditions :
        return None , None

    clauses = []
    params = []
    
    for field, value in conditions.items() :
        if isinstance(value, tuple) and len(value) == 2 :
            operator, val = value
            # 新增：如果val是NDayInterval，拼接SQL表达式
            if isinstance(val, NDayInterval):
                clauses.append(f"{field} {operator} {val}")
            # 处理IN和NOT IN运算符的特殊情况
            elif operator.upper() in ('IN', 'NOT IN') and isinstance(val, (list, tuple)):
                # 校验列表/元组中的每个元素
                validated_val = []
                for item in val:
                    validated_item = _validate_param_value(item, field)
                    validated_val.append(validated_item)
                val = validated_val
                placeholders = ', '.join(['%s'] * len(val))
                clauses.append(f"{field} {operator.upper()} ({placeholders})")
                params.extend(val)
            else:
                # 校验参数值
                validated_val = _validate_param_value(val, field)
                clauses.append(f"{field} {operator} %s")
                params.append(validated_val)
        else :
            # 校验参数值
            validated_value = _validate_param_value(value, field)
            clauses.append(f"{field} = %s")
            params.append(validated_value)
            
    where_clause = ' AND '.join(clauses)
    return where_clause , params



from ..tools.where_clause import build_where_clause

def delete(executor, table_name, conditions, commit=False, self_close=False):
    """
    通用的SQL删除执行器方法，支持动态构造WHERE子句

    :param executor: SQLExecutor 实例
    :param table_name: 表名
    :param conditions: WHERE条件，格式为字典，如 {'field1': 'value1', 'field2': 'value2'}
    :param commit: 是否自动提交
    :param self_close: 是否自动关闭连接
    :return: None
    """

    # 构造WHERE子句
    where_clause, params = build_where_clause(conditions)

    # 构造SQL语句
    sql = f'''DELETE FROM {table_name} WHERE {where_clause};'''

    # 执行SQL
    executor.execute(sql, params, commit, self_close)


# SQL工具函数

# 载入sql文件
def load_sql( sql_path ) :
    with open(sql_path, 'r', encoding='utf-8') as f:
        sql = f.read()
    return sql.strip()

# 构建SQL条件限制语句
def add_limit( column , value , column_alias = "" , add_and = True , operator = "=" ) : 
    """
    构建SQL条件限制语句，支持多种比较运算符
    
    :param column: 字段名
    :param value: 字段值，如果为"", "all", "null"则返回空字符串
    :param column_alias: 表别名，可选
    :param add_and: 是否添加AND前缀
    :param operator: 比较运算符，支持 =, !=, <>, >, >=, <, <=, LIKE, NOT LIKE, IN, NOT IN
    :return: SQL条件语句片段
    
    :example:
        >>> add_limit('status', 'active')
        "AND status = 'active'"
        >>> add_limit('age', 18, 'u', operator='>=')
        "AND u.age >= 18"
        >>> add_limit('name', '%张%', operator='LIKE')
        "AND name LIKE '%张%'"
        >>> add_limit('type', ['admin', 'user'], operator='IN')
        "AND type IN ('admin', 'user')"
    """
    if value in ("" , "all", "null") : 
        return "" 
    if column_alias != "" : 
        column_alias += "." 
    
    # 处理IN/NOT IN运算符的特殊情况
    if operator.upper() in ('IN', 'NOT IN') and isinstance(value, (list, tuple)):
        value_str = ', '.join([str(v) if isinstance(v, (int, float)) else f"'{v}'" for v in value])
        condition = f"{column_alias}{column} {operator.upper()} ({value_str})"
    else:
        # 处理数字类型，不需要加单引号
        if isinstance(value, (int, float)):
            condition = f"{column_alias}{column} {operator.upper()} {value}"
        else:
            condition = f"{column_alias}{column} {operator.upper()} '{value}'"
    
    if add_and : 
        return f"AND {condition}" 
    else : 
        return condition



"""
Tests for batch_update functionality
"""
import pytest
from lazy_mysql.utils.update.batch_update import (
    _build_complex_update_sql,
    _build_simple_update_sql,
    _check_simple_case
)
import pandas as pd


def test_complex_batch_update_parameter_order():
    """
    Test that parameters are in correct order for complex batch update
    
    The SQL should be:
    UPDATE table SET field = CASE WHEN cond1 = %s AND cond2 = %s THEN %s ...
    
    And params should be:
    (cond1_value, cond2_value, field_value, ...)
    NOT (field_value, cond1_value, cond2_value, ...)
    """
    update_list = [
        {
            'conditions': {'order_id': '1143212438', 'platform': 'Qoo10_ja', 'shop_name': 'shun_ni'},
            'fields': {'estimate_ship_at': '2026-01-22 23:59:59'}
        },
        {
            'conditions': {'order_id': '1143275182', 'platform': 'Qoo10_ja', 'shop_name': 'fortune_awaits'},
            'fields': {'estimate_ship_at': '2026-01-22 23:59:59'}
        }
    ]
    
    all_fields = ['estimate_ship_at']
    
    sql, params = _build_complex_update_sql('order_info', update_list, all_fields)
    
    # The SQL should contain WHEN conditions followed by THEN
    assert 'WHEN' in sql
    assert 'THEN' in sql
    assert 'CASE' in sql
    
    # Check parameter order
    # For the first WHEN clause, the parameters should be:
    # order_id, platform, shop_name (WHERE params), then estimate_ship_at (THEN value)
    
    # The params should start with the SET clause parameters
    # For first record: order_id='1143212438', platform='Qoo10_ja', shop_name='shun_ni', 
    # THEN estimate_ship_at='2026-01-22 23:59:59'
    
    # Extract SET clause params (before WHERE clause params)
    # Expected order in SET clause for each CASE WHEN:
    # WHERE params (order_id, platform, shop_name), THEN value (estimate_ship_at)
    
    # Total params: 
    # SET clause: 2 records * (3 WHERE params + 1 THEN value) = 8 params
    # WHERE clause: 2 records * 3 WHERE params = 6 params
    # Total: 14 params
    assert len(params) == 14
    
    # The first 4 params should be for the first CASE WHEN in SET clause:
    # order_id, platform, shop_name, estimate_ship_at
    assert params[0] == '1143212438', f"Expected order_id first, got {params[0]}"
    assert params[1] == 'Qoo10_ja', f"Expected platform second, got {params[1]}"
    assert params[2] == 'shun_ni', f"Expected shop_name third, got {params[2]}"
    assert params[3] == '2026-01-22 23:59:59', f"Expected estimate_ship_at fourth, got {params[3]}"
    
    # The next 4 params should be for the second CASE WHEN in SET clause:
    # order_id, platform, shop_name, estimate_ship_at
    assert params[4] == '1143275182'
    assert params[5] == 'Qoo10_ja'
    assert params[6] == 'fortune_awaits'
    assert params[7] == '2026-01-22 23:59:59'
    
    # The last 6 params are for the WHERE clause (2 records * 3 params each)
    assert params[8] == '1143212438'
    assert params[9] == 'Qoo10_ja'
    assert params[10] == 'shun_ni'
    assert params[11] == '1143275182'
    assert params[12] == 'Qoo10_ja'
    assert params[13] == 'fortune_awaits'


def test_simple_batch_update_parameter_order():
    """
    Test that parameters are in correct order for simple batch update (single condition field)
    """
    update_list = [
        {'conditions': {'id': 1}, 'fields': {'name': 'Alice', 'age': 25}},
        {'conditions': {'id': 2}, 'fields': {'name': 'Bob', 'age': 30}}
    ]
    
    all_fields = ['name', 'age']
    key_field = 'id'
    df_conditions = pd.DataFrame([item['conditions'] for item in update_list])
    
    sql, params = _build_simple_update_sql('users', update_list, all_fields, key_field, df_conditions)
    
    # SQL should use CASE key_field WHEN pattern
    assert 'CASE id' in sql
    assert 'WHEN' in sql
    assert 'THEN' in sql
    
    # Parameters should be in order:
    # SET clause: (key1, value1, key2, value2, ...) for each field
    # WHERE clause: (key1, key2, ...)
    
    # For name field: id=1, name='Alice', id=2, name='Bob'
    # For age field: id=1, age=25, id=2, age=30
    # WHERE IN: id=1, id=2
    
    # Total: 2 + 2 + 2 + 2 + 2 = 10 params
    assert len(params) == 10
    
    # First field (name) params
    assert params[0] == 1  # id for first WHEN
    assert params[1] == 'Alice'  # name value
    assert params[2] == 2  # id for second WHEN
    assert params[3] == 'Bob'  # name value
    
    # Second field (age) params
    assert params[4] == 1  # id for first WHEN
    assert params[5] == 25  # age value
    assert params[6] == 2  # id for second WHEN
    assert params[7] == 30  # age value
    
    # WHERE IN clause params
    assert params[8] == 1
    assert params[9] == 2


def test_check_simple_case():
    """Test the logic for detecting simple case (single condition field)"""
    # Single condition field - should be simple case
    df_simple = pd.DataFrame([{'id': 1}, {'id': 2}])
    is_simple, key_field = _check_simple_case(df_simple)
    assert is_simple is True
    assert key_field == 'id'
    
    # Multiple condition fields - should not be simple case
    df_complex = pd.DataFrame([
        {'id': 1, 'type': 'user'},
        {'id': 2, 'type': 'admin'}
    ])
    is_simple, key_field = _check_simple_case(df_complex)
    assert is_simple is False
    assert key_field is None
    
    # Single condition field with tuple values - should not be simple case
    df_tuple = pd.DataFrame([{'id': ('>', 1)}, {'id': ('>', 2)}])
    is_simple, key_field = _check_simple_case(df_tuple)
    assert is_simple is False
    assert key_field is None


if __name__ == '__main__':
    pytest.main([__file__, '-v'])
